
//var org.greenstone.meandre.audio.util.MixType = {

var MixType = {
    LeftOnly:     { value: 0, name: "Left only" },
    RightOnly:    { value: 1, name: "Right only" },
    LeftRightAdd: { value: 2, name: "Left and Right add" },
    LeftRightSub: { value: 3, name: "Left and Right subtract" }
};


var AudioMixer = function(numSamples,channels) {

    this.getset = {};

    this.numSamples = numSamples;
    this.channels = 2;

}

AudioMixer.inherits(Component);

//----------------------------- OVERVIEW -----------------------------------------------------

AudioMixer.staticfield('Component', {

    creator: "David Bainbridge (based on AudioFileRead by Kris West)", 

    description : "Overview: Takes a Signal object with file location metadata as input, "
	+ "reads in the corresponding audio file and streams out frames of data from that file."
	+ "Detailed Description: This module reads in audio files in a number of formats,"
	+ "enframes the data and streams out the frames. Mixing options include: LeftOnly, RightOnly, "
	+ "LeftRightAdd, LeftRightSub."
	+ "The frame size, overlap percentage and sample rate of the of the audio frames are appended to"
	+ "the metadata of the Signal object, which is \'passed through\' for later (optional) use but other components. "
	+ "The \"Read Frame Size\" parameter can be used to control how often data is read from "
	+ "disk."
	+ "Currently supported audio file formats include MP3 (.mp3) and Wave (.wav).",

    name: "Audio Signal Mixer", 

    tags: "signal reader",

    // dependency: [ "foundry-abstracts.jar", "seasr-commons-1.4.9.jar" ]
});

//------------------------------ INPUTS ------------------------------------------------------

AudioMixer.staticfield('ComponentInputArray', [
/*
    {
	name: "signal",
	description: "A Signal object with \"fileLocation\" metadata, indicating the " 
	    + "location of the corresponding audio file."
    }
*/

    {
	name: "rawInput",
	description: "The raw float32 input from the underlying (decompressed if required) audio file"
    }


]);


//------------------------------ PROPERTIES --------------------------------------------------

AudioMixer.staticfield('ComponentPropertyArray', [
/*
    {
	name: "outputFrameLength",
	defaultValue: "0.023219954648526078",
	description: "The length of audio frames to output, measured in seconds. This "
	    + "will be multiplied by the samplerate of the audio file and "
	    + "rounded to calculate the length in samples of the output "
	    + "vectors. This ensures that constant time resolution can be "
	    + "maintained at different sample rates. The default setting (0.023219...) "
	    + "produces 1024 sample frames at 44.1kHz and 512 at 22.05kHz."
    },

    {
	name: "overlapPercent",
	defaultValue: "0.5", 
	description: "The percentage to overlap concurrent output frames by (0.0-1.0)"
    },
    
    {
	name: "mixType",
	defaultValue: "LeftRightAdd", 
	description: "Controls how the Signal in should be mix.  Assuming a stero "
	    + "input, then choices are: LeftOnly, RightOnly, LeftRightAdd, LeftRightSub."
	    + "In the case of mono input, the signal is passed straight through unchanged."
    },


    {
	name: "readFrameSize",
	defaultValue: "102400", 
	description: "Controls frame size of data read in by this module. If possible, "
	    + "this frame size will be used by the AudioReader responsible for "
	    + "the file otherwise buffering will be used to achieve the correct "
	    + "framesize. This parameter is used to control how often data is "
	    + "read from disk."
    },
    
    
    {
	name: "limitAudioFrames",
	defaultValue: "false", 
	description: "Controls whether the number of frames that are output is limited or not."
    },
    
    
    {
	name: "maxOutFrames",
	defaultValue: "10000", 
	description: "The maxium number of frames to output."
    }
*/
]);


//------------------------------ OUTPUTS -----------------------------------------------------

AudioMixer.staticfield('ComponentOutputArray', [

    {
	name: "mixedFrames",
	description: "Audio frames that are some combination of left and/or right channels "
	    + "as set by the mixType property"
    },
/*
    {
	name: "signal",
	description: "The Signal Object used to read the file, with"
	    + " \"frameSize\", \"overlapSize\" and \"sampleRate\" metadata set."
    }
*/

]);

AudioMixer.prototype.defaults = function() {

    /** enum representing how the left and right audio in channels are to be treated
     */
    getset.mixType = MixType.LeftRightAdd;

    /** The length of audio frames output, measured in seconds. */
    getset.outputFrameLength = (512.0 / 22050.0);

    /** The desired read frame size. This parameter can be used to adjust the number
     * of disk accesses per output frame.
     */
    getset.readFrameSize = 1024;

    /** The percentage to overlap concurrent frames.
     */
    getset.overlapPercent = 0.5;

    /** The maximum number of frames to output.
     */
    getset.maxOutFrames = 10000;
    /** Determines whether number of frames is limited
     */
    getset.limitAudioFrames = false;

}


AudioMixer.prototype.initializeCallBack = function(ccp) {

    // typeof(ccp) => ComponentContextProperties

    // Put these into ComponentPropertyArray

    this.concatLimit = 8;
    this.concatCount = 0;

    this.concatSignal = new Float32Array(this.numSamples * this.concatLimit); // mono

}


AudioMixer.prototype.executeCallBack = function(cc) {
    
    // typeof(cc) => ComponentContext
    
    // get the "[D" input

    //console.log("*** AudioMixer::executeCallBack(): away to request 'rawInput'");

    var fb = cc.getDataComponentFromInput("rawInput");
/*
    var mess = "";
    for (var i=0; i<fb.length; i+=200) {
	mess += fb[i] + ", ";
    }
    console.log("**** AudioMixer: data samples = " + mess);
*/

    // Mix number of channels down to one channel

    var p =0;
    for (var i=0; i<this.numSamples; i++ ) {
	
	var sv  = 0;
	for (var c=0; c<this.channels; c++) {
	    sv += fb[p];
	    p++;
	}
	// signal[i] = sv / channels;
	this.concatSignal[(this.concatCount*this.numSamples)+i] = sv / this.channels;
    }
    
    this.concatCount++;

    //console.log("*** concatCount = " + this.concatCount);
    //console.log("*** concatLimit = " + this.concatLimit);

    if (this.concatCount==this.concatLimit) {
	// send this mono-mixed concatenated array on its way

	this.concatCount=0;
        cc.pushDataComponentToOutput("mixedFrames", this.concatSignal);
    }
}


AudioMixer.prototype.disposeCallBack = function(ccp) {

    // typeof(ccp) => ComponentContextProperties

    this.concatSignal = null;

}





