
var JsMadProcessing = function(audio_id,active_workflow) {

    this.audio_id = audio_id;
    this.active_workflow = active_workflow || null;
  
    var url = $('#'+audio_id).attr("href");
    console.info("JsMadProcessing::JsMadProcessing() retrieving over AJAX: " + url);

    var stream = new Mad.AjaxStream(url);

    var that = this;

    var channels = 2; // stereo
    var rate = 44100; // 44.1 Khz
    //var rate = 48000; // 44.1 Khz
    //var bufsize = 65536 * 4096; 
    //var bufsize = 65536 * 1024; 
    //var bufsize = 2048;
    var bufsize = 1024 * 8;

    //this.dev = Sink(function(buffer) {that.refill(buffer)}, channels, bufsize, rate);
    //console.log("*** got sample rate of: " + this.dev.sampleRate);


    this.playBufferStore = new Array();

    stream.requestAbsolute(1 * 1024, function() { that.streamCallback(stream) });

}




JsMadProcessing.prototype.streamCallback = function(stream) {


    var mp3 = new Mad.MP3File(stream);
    var mpeg = mp3.getMpegStream();
        
    //var howmuch = 10000; 
    var ss1 = Date.now();

    var synth = new Mad.Synth();
    var frame = new Mad.Frame();
	
    frame = Mad.Frame.decode(frame, mpeg); // decode first frame to establish num channels and samplerate
    var num_channels = frame.header.nchannels();
    var sampleRate = frame.header.samplerate;

    console.info("JsMadProcessing::streamCallback() decoding URL retrieved MP3 with " + num_channels + " channels, samplerate = " + sampleRate);


    var that = this;

    var numSamples = 0;

    synth.frame(frame);    
    var prev_synth_frame_len = synth.pcm.samples[0].length; 
    console.info("JsMadProcesing.streamCallback(): synth frame length (initial frame) = " + prev_synth_frame_len);

    var i=1; // already processed one frame

    var overflowBuffer = null;

    console.log("*** main-jsmad.js::stream_callback(): hardwiring workflow buffer size to 2048");

    var speedup_batch_limit;
    var speedup_pause;
    
    if (BrowserDetect.browser == "Firefox") {
	speedup_batch_limit = 100;
	speedup_pause = 0;
    }
    else {
	// e.g. Chrome 
	// => the following values played more nicely with letting DSP playback operate straightaway 
	//    i.e. when jsmad is still working its way through the frames
	
	speedup_batch_limit = 10;
	speedup_pause = 2;

	//speedup_batch_limit = 100;
	//speedup_pause = 0;

	console.log("*** Running JsMadProcessing with: batch limit = " + speedup_batch_limit 
		    + ", pause = " + speedup_pause + " msecs");
    }


    (function() {

	var more_frames = 1;
	try {

	    var speedup_i = 0;

	    while (speedup_i < speedup_batch_limit) { // Decode 100 frames at a time, before setTimeout break

		frame = Mad.Frame.decode(frame, mpeg);
		synth.frame(frame);
	    
		var synth_frame_len = synth.pcm.samples[0].length; 

		var playBuffer = null;
		var pbp = 0; // pbp = play buffer pos

		if (overflowBuffer != null) {
		    playBuffer = new Float32Array(overflowBuffer.length+(synth_frame_len*num_channels));
		    playBuffer.set(overflowBuffer);
		    pbp = overflowBuffer.length;
		    overflowBuffer = null;
		}
		else {
		    playBuffer = new Float32Array(synth_frame_len*num_channels);
		    pbp = 0;
		}

				
		for (var sfp=0; sfp<synth_frame_len; sfp++) {
		    for (var c=0; c<num_channels; c++) {
			playBuffer[pbp] = synth.pcm.samples[c][sfp];
			pbp++;
		    }
		}
		
		numSamples += synth_frame_len;

		// uncomment the following line if you want to hear speeded-up playback as it works through the audio file
		//var written = pannedPlayback.mozWriteAudio(playBuffer);


		if (synth_frame_len != prev_synth_frame_len) {
		    console.log("*** new synth frame len (frame num = " + i + "): len = " + synth_frame_len);
		    prev_synth_frame_len = synth_frame_len;
		}

		if (that.active_workflow != null) {
		    // Send the data into the Javascript DSP workflow

		    // Note: workflow geared to 2048 samples at a time

		    var wi = 0;

		    var playBufferLen = playBuffer.length;
		    while (wi<playBuffer.length) {

			if (wi+2048<=playBufferLen) {			    
			    var workflowBuffer = playBuffer.subarray(wi,wi+2048);
			    that.active_workflow.pumpDataHead("rawInput",workflowBuffer);
			    that.playBufferStore.push(workflowBuffer);
			}
			else {
			    // tail end => set up in overflow buffer
			    var workflowBufferTail = playBuffer.subarray(wi);

			    overflowBuffer = new Float32Array(playBufferLen-wi);
			    overflowBuffer.set(workflowBufferTail);

			}
			wi += 2048;
		    }


		    // Old approach: pump data in, in frames that are set by webAudioContext
		    //that.active_workflow.pumpDataHead("rawInput",playBuffer);
		}

		speedup_i++;
		i++;
	    }
	    
	}
	catch(err) {
	    if (err == "Header decoding failed") {
		console.info("JsMadProcessing::streamCallack() decoded " + i + " frames");
	    }
	    else {
		console.error(err);
	    }
	    more_frames = 0;
	}

	if (more_frames) {
	    setTimeout(arguments.callee,speedup_pause); // even a delay of '0' gives the browser thread a break
	}
	else {
	    var ss2 = Date.now();

	    var song_duration = numSamples / sampleRate;
	    var speedup_duration = (ss2 - ss1)/1000;

	    var speedup_factor = song_duration / speedup_duration;
	    //console.log("Computed song duration: " + song_duration + " secs");
	    console.info("JsMad processing done in " + speedup_duration + " secs (a speed factor of " + speedup_factor + ")");
	}
    	
    })();
	
}


JsMadProcessing.prototype.refill = function(buffer) 
{
    console.log("refill called, asking for size = " + buffer.length);

    var buffer_len = buffer.length;
    var buffer_pos = 0;

    while ((this.playBufferStore.length>0) && (buffer_pos<buffer_len)) {

	var playBuffer = this.playBufferStore.shift();
	buffer.set(playBuffer,buffer_pos);

	buffer_pos += playBuffer.length;
    }


}
